import enum
import functools
import re

import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    IMAGEDATA_ELEMENT,
    NDARRAY_ELEMENT,
    NODE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    UINT64_ELEMENT,
    ChannelMetaData,
    Dims,
    Encoding,
    Epochstamp,
    Hash,
    ImageData,
    MetricPrefix,
    Schema,
    Timestamp,
    Trainstamp,
    Unit,
)
from TrainMatcher import TrainMatcher

from . import geom_utils, scenes, utils
from ._version import version as deviceVersion


assembled_schema = Schema()
(
    NODE_ELEMENT(assembled_schema).key("image").commit(),

    NDARRAY_ELEMENT(assembled_schema).key("image.data").commit(),

    UINT64_ELEMENT(assembled_schema).key("trainId").readOnly().commit(),
)

preview_schema = Schema()
(
    NODE_ELEMENT(preview_schema).key("image").commit(),

    IMAGEDATA_ELEMENT(preview_schema).key("image.data").commit(),

    UINT64_ELEMENT(preview_schema).key("trainId").readOnly().commit(),
)

xtdf_source_re = re.compile(r".*\/DET\/(\d+)CH0:xtdf")
daq_source_re = re.compile(r".*\/DET\/.*?(\d+):daqOutput")


class BridgeOutputOptions(enum.Enum):
    MATCHED = "matched"
    ASSEMBLED = "assembled"
    PREVIEW = "preview"


# TODO: merge scene with TrainMatcher's nice overview
@KARABO_CLASSINFO("DetectorAssembler", deviceVersion)
class DetectorAssembler(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("availableScenes")
            .setNewDefaultValue(["overview", "trainMatcherScene"])
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("timeOfFlight")
            .displayedName("Time of flight")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(1000)
            .info("Time of flight exceeding 1 s")
            .needsAcknowledging(False)
            .commit(),

            STRING_ELEMENT(expected)
            .key("pathToStack")
            .assignmentOptional()
            .defaultValue("image.data")
            .commit(),

            OUTPUT_CHANNEL(expected)
            .key("assembledOutput")
            .dataSchema(assembled_schema)
            .commit(),

            NODE_ELEMENT(expected)
            .key("preview")
            .description(
                "The preview output is intended for Karabo GUI previews. It differs "
                "from the main output in that it is rate throttled, can be "
                "downsampled, and is given the ImageData type for use within Karabo."
            )
            .commit(),

            OUTPUT_CHANNEL(expected)
            .key("preview.output")
            .dataSchema(preview_schema)
            .description("See description of parent node, 'preview'.")
            .commit(),

            UINT32_ELEMENT(expected)
            .key("preview.downsamplingFactor")
            .description(
                "If greater than 1, the assembled image will be downsampled by this "
                "factor in x and y dimensions before sending. This is only to save "
                "bandwidth in case GUI updates start lagging."
            )
            .assignmentOptional()
            .defaultValue(1)
            .options("1,2,4,8")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("preview.downsamplingFunction")
            .description("Reduction function used during downsampling.")
            .assignmentOptional()
            .defaultValue("nanmax")
            .options("nanmax,nanmean,nanmin,nanmedian")
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("preview.maxRate")
            .displayedName("Max rate")
            .description(
                "Preview output is throttled to (at most) this speed. New trains "
                "matched 'too soon' get dropped here (instead of sending to be dropped "
                "by GUI)."
            )
            .unit(Unit.HERTZ)
            .assignmentOptional()
            .defaultValue(2)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("outputForBridgeOutput")
            .displayedName("Bridge output")
            .description(
                "The bridge can send the same data as any one of the three Karabo "
                "output channels. If this setting is 'matched', it will follow "
                "'output', if 'assembled', it will follow 'assembledOutput' and if "
                "'preview', it will follow 'preview.output'."
            )
            .options(",".join(option.value for option in BridgeOutputOptions))
            .assignmentOptional()
            .defaultValue("matched")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("geometryDevice")
            .displayedName("Geometry device")
            .description(
                "The the device which will provide geometry. The device is expected to "
                "have current geometry as a string (for now using custom non-robust "
                "serialization scheme) element named serializedGeometry."
            )
            .assignmentMandatory()
            .commit(),
        )

    def __init__(self, conf):
        super().__init__(conf)
        self.info.merge(Hash("timeOfFlight", 0))

    def initialization(self):
        super().initialization()

        # TODO: match inside device, fill multiple independent buffers

        self._throttler = utils.SkippingThrottler(1 / self.get("preview.maxRate"))
        self._path_to_stack = self.get("pathToStack")
        self._geometry = None
        self._stack_input_buffer = None

        self.KARABO_SLOT(self.requestScene)

        geometry_device = self.get("geometryDevice")
        client = self.remote()
        try:
            initial_geometry = client.get(geometry_device, "serializedGeometry")
        except RuntimeError:
            self.log.WARN(
                "Failed to get initial geometry, maybe geometry device is down"
            )
        else:
            self.log.INFO("Got geometry immediately after init :D")
            self._receive_geometry(
                geometry_device,
                Hash("pickledGeometry", initial_geometry),
            )

        self.remote().registerDeviceMonitor(geometry_device, self._receive_geometry)

        self.assembled_output = self.signalSlotable.getOutputChannel("assembledOutput")
        self.preview_output = self.signalSlotable.getOutputChannel("preview.output")
        self.start()

    def requestScene(self, params):
        # TODO: unify with TrainMatcher overview
        scene_name = params.get("name", default="")
        if scene_name == "overview":
            payload = Hash("name", scene_name, "success", True)
            payload["data"] = scenes.detector_assembler_overview(
                device_id=self.getInstanceId()
            )
            self.reply(
                Hash(
                    "type",
                    "deviceScene",
                    "origin",
                    self.getInstanceId(),
                    "payload",
                    payload,
                )
            )
        elif scene_name == "trainMatcherScene":
            params["name"] = "scene"
            return super().requestScene(params)

    def _receive_geometry(self, device_id, config):
        if not config.has("serializedGeometry"):
            return
        self.log.INFO(f"Found geometry on {device_id}")
        serialized_geometry = config["serializedGeometry"]
        if len(serialized_geometry) == 0:
            self.log.INFO("New geometry empty, will ignore update.")
            return
        self._geometry = geom_utils.deserialize_geometry(serialized_geometry)
        # TODO: allow multiple memory cells (extra geom notion of extra dimensions)
        self._stack_input_buffer = np.zeros(self._geometry.expected_data_shape)

    def on_matched_data(self, train_id, sources):
        if self._geometry is None:
            self.log.WARN("Have not received a geometry yet, will not send anything")
            return

        my_timestamp = Timestamp(Epochstamp(), Trainstamp(train_id))
        my_source = self.getInstanceId()
        bridge_output_choice = BridgeOutputOptions(
            self.unsafe_get("outputForBridgeOutput")
        )

        module_indices_unfilled = set(range(self._stack_input_buffer.shape[0]))
        earliest_source_timestamp = float("inf")
        for source, (data, source_timestamp) in sources.items():
            # regular TrainMatcher output
            self.output.write(data, ChannelMetaData(source, source_timestamp))
            if bridge_output_choice is BridgeOutputOptions.MATCHED:
                self.zmq_output.write(source, data, source_timestamp)

            # prepare for assembly
            # TODO: handle failure to "parse" source, get data out
            module_index = self._source_to_index(source)
            self._stack_input_buffer[module_index] = np.squeeze(
                data.get(self._path_to_stack)
            )
            module_indices_unfilled.discard(module_index)
            earliest_source_timestamp = min(
                earliest_source_timestamp, source_timestamp.toTimestamp()
            )

        self.output.update()
        if bridge_output_choice is BridgeOutputOptions.MATCHED:
            self.zmq_output.update()

        for module_index in module_indices_unfilled:
            self._stack_input_buffer[module_index].fill(0)
            # TODO: configurable treatment of missing modules

        # TODO: reusable output buffer to save on allocation
        assembled, _ = self._geometry.position_modules_fast(self._stack_input_buffer)

        # TODO: optionally include control data
        output_hash = Hash(
            "image.data",
            assembled,
            "trainId",
            train_id,
        )
        output_metadata = ChannelMetaData(my_source, my_timestamp)
        self.assembled_output.write(output_hash, output_metadata)
        self.assembled_output.update()
        if bridge_output_choice is BridgeOutputOptions.ASSEMBLED:
            self.zmq_output.write(my_source, output_hash, my_timestamp)
            self.zmq_output.update()

        if self._throttler.test_and_set():
            downsampling_factor = self.unsafe_get("preview.downsamplingFactor")
            if downsampling_factor > 1:
                assembled = downsample_2d(
                    assembled,
                    downsampling_factor,
                    reduction_fun=getattr(
                        np, self.unsafe_get("preview.downsamplingFunction")
                    ),
                )
            output_hash = Hash(
                "image.data",
                ImageData(
                    # TODO: get around this being mirrored...
                    assembled.astype(np.int32)[::-1, ::-1],
                    Dims(*assembled.shape),
                    Encoding.GRAY,
                    bitsPerPixel=32,
                ),
                "trainId",
                train_id,
            )
            self.preview_output.write(
                output_hash,
                output_metadata,
            )
            self.preview_output.update()
            if bridge_output_choice is BridgeOutputOptions.PREVIEW:
                self.zmq_output.write(
                    my_source,
                    output_hash,
                    my_timestamp,
                )
                self.zmq_output.update()

        self.info["timeOfFlight"] = (
            Timestamp().toTimestamp() - earliest_source_timestamp
        ) * 1000
        self.info["sent"] += 1
        self.info["trainId"] = train_id
        self.rate_out.update()

    def on_new_data(self, channel, data, meta):
        super().on_new_data(channel, data, meta)
        ...
        # TODO: start filling buffer early

    @functools.lru_cache()
    def _source_to_index(self, source):
        # note: cache means warning only shows up once (also not performance-critical)
        # TODO: allow user to inspect, modify the mapping

        match = xtdf_source_re.match(source)
        if match is not None:
            return int(match.group(1))

        match = daq_source_re.match(source)
        if match is not None:
            return int(match.group(1)) - 1

        self.log.WARN(f"Couldn't figure out index for source {source}")
        return 0


def downsample_2d(arr, factor, reduction_fun=np.nanmax):
    """Generalization of downsampling from FemDataAssembler

    Expects first two dimensions of arr to be multiple of 2 ** factor
    Useful if you're sitting at home and ssh connection is slow to get full-resolution
    previews."""

    for i in range(factor // 2):
        arr = reduction_fun(
            (
                arr[:-1:2],
                arr[1::2],
            ),
            axis=0,
        )
        arr = reduction_fun(
            (
                arr[:, :-1:2],
                arr[:, 1::2],
            ),
            axis=0,
        )
    return arr


# forward-compatible unsafe_get proposed by @haufs
if not hasattr(DetectorAssembler, "unsafe_get"):

    def unsafe_get(self, key):
        """See base_correction.py"""
        return self._parameters.get(key)

    setattr(DetectorAssembler, "unsafe_get", unsafe_get)
